@@ -181,6 +181,12 @@ nsSVGPathGeometryFrame::GetType() const
   return nsLayoutAtoms::svgPathGeometryFrame;
 }
 
+PRBool
+nsSVGPathGeometryFrame::IsFrameOfType(PRUint32 aFlags) const
+{
+  return !(aFlags & ~nsIFrame::eSVG);
+}
+
 //----------------------------------------------------------------------
 // nsISVGChildFrame methods
 
@@ -246,6 +252,10 @@ nsSVGPathGeometryFrame::PaintSVG(nsISVGRendererCanvas* canvas, const nsRect& dir
     GetMarkerFrames(&markerStart, &markerMid, &markerEnd);
 
     if (markerEnd || markerMid || markerStart) {
+      // need to set this up with the first draw
+      if (!mMarkerRegion)
+        mMarkerRegion = GetCoveredRegion();
+
       float strokeWidth;
       GetStrokeWidth(&strokeWidth);
       
@@ -309,11 +319,11 @@ nsSVGPathGeometryFrame::GetFrameForPointSVG(float x, float y, nsIFrame** hit)
 NS_IMETHODIMP_(already_AddRefed<nsISVGRendererRegion>)
 nsSVGPathGeometryFrame::GetCoveredRegion()
 {
-  nsISVGRendererRegion *region = nsnull;
   if (!GetGeometry())
-    return region;
+    return nsnull;
 
-  GetGeometry()->GetCoveredRegion(&region);
+  nsCOMPtr<nsISVGRendererRegion> region;
+  GetGeometry()->GetCoveredRegion(getter_AddRefs(region));
 
   nsISVGMarkable *markable;
   CallQueryInterface(this, &markable);
@@ -322,44 +332,57 @@ nsSVGPathGeometryFrame::GetCoveredRegion()
     nsSVGMarkerFrame *markerEnd, *markerMid, *markerStart;
     GetMarkerFrames(&markerStart, &markerMid, &markerEnd);
 
-    if (!markerEnd && !markerMid && !markerStart)
-      return region;
+    if (markerEnd || markerMid || markerStart) {
+      float strokeWidth;
+      GetStrokeWidth(&strokeWidth);
 
-    float strokeWidth;
-    GetStrokeWidth(&strokeWidth);
+      nsVoidArray marks;
+      markable->GetMarkPoints(&marks);
 
-    nsVoidArray marks;
-    markable->GetMarkPoints(&marks);
+      PRUint32 num = marks.Count();
 
-    PRUint32 num = marks.Count();
-
-    if (markerStart) {
-      nsCOMPtr<nsISVGRendererRegion> mark;
-      mark = markerStart->RegionMark(this, (nsSVGMark *)marks[0], strokeWidth);
-
-      nsCOMPtr<nsISVGRendererRegion> tmp = dont_AddRef(region);
-      mark->Combine(tmp, &region);
-    }
-
-    if (markerMid)
-      for (PRUint32 i = 1; i < num - 1; i++) {
+      if (num && markerStart) {
         nsCOMPtr<nsISVGRendererRegion> mark;
-        mark = markerMid->RegionMark(this, (nsSVGMark *)marks[i], strokeWidth);
+        mark = markerStart->RegionMark(this, (nsSVGMark *)marks[0], strokeWidth);
 
-        nsCOMPtr<nsISVGRendererRegion> tmp = dont_AddRef(region);
-        mark->Combine(tmp, &region);
+        if (!region) {
+          region = mark;
+        } else if (mark) {
+          nsCOMPtr<nsISVGRendererRegion> tmp = region;
+          mark->Combine(tmp, getter_AddRefs(region));
+        }
       }
 
-    if (markerEnd) {
-      nsCOMPtr<nsISVGRendererRegion> mark;
-      mark = markerEnd->RegionMark(this, (nsSVGMark *)marks[num-1], strokeWidth);
+      if (num && markerMid)
+        for (PRUint32 i = 1; i < num - 1; i++) {
+          nsCOMPtr<nsISVGRendererRegion> mark;
+          mark = markerMid->RegionMark(this, (nsSVGMark *)marks[i], strokeWidth);
 
-      nsCOMPtr<nsISVGRendererRegion> tmp = dont_AddRef(region);
-      mark->Combine(tmp, &region);
+          if (!region) {
+            region = mark;
+          } else if (mark) {
+            nsCOMPtr<nsISVGRendererRegion> tmp = region;
+            mark->Combine(tmp, getter_AddRefs(region));
+          }
+        }
+
+      if (num && markerEnd) {
+        nsCOMPtr<nsISVGRendererRegion> mark;
+        mark = markerEnd->RegionMark(this, (nsSVGMark *)marks[num-1], strokeWidth);
+
+        if (!region) {
+          region = mark;
+        } else if (mark) {
+          nsCOMPtr<nsISVGRendererRegion> tmp = region;
+          mark->Combine(tmp, getter_AddRefs(region));
+        }
+      }
     }
   }
 
-  return region;
+  nsISVGRendererRegion *retval = nsnull;
+  region.swap(retval);
+  return retval;
 }
 
 NS_IMETHODIMP
@@ -388,17 +411,9 @@ nsSVGPathGeometryFrame::NotifyRedrawSuspended()
 NS_IMETHODIMP
 nsSVGPathGeometryFrame::NotifyRedrawUnsuspended()
 {
-  if (mUpdateFlags != 0) {
-    nsCOMPtr<nsISVGRendererRegion> dirty_region;
-    if (GetGeometry())
-      GetGeometry()->Update(mUpdateFlags, getter_AddRefs(dirty_region));
-    if (dirty_region) {
-      nsISVGOuterSVGFrame* outerSVGFrame = GetOuterSVGFrame();
-      if (outerSVGFrame)
-        outerSVGFrame->InvalidateRegion(dirty_region, PR_TRUE);
-    }
-    mUpdateFlags = 0;
-  }
+  if (mUpdateFlags != 0)
+    UpdateGraphic(0);
+
   return NS_OK;
 }
 
@@ -537,18 +552,34 @@ nsSVGPathGeometryFrame::GetStrokeDashArray(float **arr, PRUint32 *count)
 {
   const nsStyleCoord *dasharray = GetStyleSVG()->mStrokeDasharray;
   nsPresContext *presContext = nsSVGPathGeometryFrameBase::GetPresContext();
+  float totalLength = 0.0f;
 
   *count = GetStyleSVG()->mStrokeDasharrayLength;
+  *arr = nsnull;
 
   if (*count) {
     *arr = (float *) nsMemory::Alloc(*count * sizeof(float));
     if (*arr) {
-      for (PRUint32 i = 0; i < *count; i++)
+      for (PRUint32 i = 0; i < *count; i++) {
         (*arr)[i] = nsSVGUtils::CoordToFloat(presContext, mContent, dasharray[i]);
+        if ((*arr)[i] < 0.0f) {
+          nsMemory::Free(*arr);
+          *count = 0;
+          *arr = nsnull;
+          return NS_OK;
+        }
+        totalLength += (*arr)[i];
+      }
     } else {
       *count = 0;
+      *arr = nsnull;
       return NS_ERROR_OUT_OF_MEMORY;
     }
+
+    if (totalLength == 0.0f) {
+      nsMemory::Free(*arr);
+      *count = 0;
+    }
   }
 
   return NS_OK;
@@ -617,7 +648,13 @@ nsSVGPathGeometryFrame::GetClipRule(PRUint16 *aClipRule)
 NS_IMETHODIMP
 nsSVGPathGeometryFrame::GetStrokePaintType(PRUint16 *aStrokePaintType)
 {
-  *aStrokePaintType = GetStyleSVG()->mStroke.mType;
+  float strokeWidth;
+  GetStrokeWidth(&strokeWidth);
+
+  // cairo will stop rendering if stroke-width is less than or equal to zero
+  *aStrokePaintType = strokeWidth <= 0 ?
+                      nsISVGGeometrySource::PAINT_TYPE_NONE :
+                      GetStyleSVG()->mStroke.mType;
   return NS_OK;
 }
 
@@ -848,11 +885,34 @@ void nsSVGPathGeometryFrame::UpdateGraphic(PRUint32 flags)
       PRBool painting;
       vm->IsPainting(painting);
 
-      if (!painting)
+      if (!painting) {
+        if (mMarkerRegion) {
+          outerSVGFrame->InvalidateRegion(mMarkerRegion, PR_TRUE);
+          mMarkerRegion = nsnull;
+        }
+        
+        nsISVGMarkable *markable;
+        CallQueryInterface(this, &markable);
+        
+        if (markable) {
+          nsSVGMarkerFrame *markerEnd, *markerMid, *markerStart;
+          GetMarkerFrames(&markerStart, &markerMid, &markerEnd);
+          
+          if (markerEnd || markerMid || markerStart) {
+            mMarkerRegion = GetCoveredRegion();
+            if (mMarkerRegion) {
+              outerSVGFrame->InvalidateRegion(mMarkerRegion, PR_TRUE);
+              mUpdateFlags = 0;
+            }
+            return;
+          }
+        }
+
         outerSVGFrame->InvalidateRegion(dirty_region, PR_TRUE);
+      }
     }
     mUpdateFlags = 0;
-  }  
+  }
 }
 
 nsISVGOuterSVGFrame *
